
# THIS FILE IS GENERATED FROM HIFIVE SETUP.PY
short_version = '1.4.0'
version = '1.4.0'
full_version = '1.4.0'
git_revision = '377f6160c01449f62f7682191a5b7a83aa35f2ff'
release = True

if not release:
    version = full_version
